import abc
import logging
from gmmmc.posterior import GMMPosteriorTarget
import numpy as np

class MonteCarloBase:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def sample(self, X, n_samples):
        return

class MarkovChain(MonteCarloBase):

    def __init__(self, proposal, prior, initial_gmm):
        """
        Base algorithm for Markov Chain Monte Carlo

        Parameters
        ----------
        proposal : Proposal Object
            The proposal function to be used.
        prior : GMMPrior Object
            The prior distribution on the parameters of the GMM.
        initial_gmm : GMM Object
            The starting point for the MCMC simulation.
        """
        self.proposal = proposal
        self.target = GMMPosteriorTarget(prior, beta=1.0)
        self.initial_gmm = initial_gmm

    def sample(self, X, n_samples, n_jobs=1):
        """
        Sample from the posterior distribution of the GMM parameters

        Parameters
        ----------
        X : 2-D array_like of shape (n_feature_vectors, n_features)
            Feature vectors used as the data for the underlying model.
        n_samples : int
            Number of Monte Carlo samples to be drawn from the posterior distribution.
        n_jobs : int
            Number of cpu cores to utilise during the Monte Carlo simulation.

        Returns
        -------
        samples : List of GMM Objects
            Returns a list of GMMs (GMM parameters) which are the samples drawn from the distribution.

        Notes
        -----
        It is beneficial to discard a number of samples from the beginning of the chain (burn-in) as well as
        to use only every nth sample (lag) to reduce the correlation between succesive samples of the posterior.
        """
        samples = []
        current_gmm = self.initial_gmm
        for run in xrange(n_samples):
            logging.info('Run: {0}'.format(run))
            current_gmm = self.proposal.propose(X, current_gmm, self.target, n_jobs)
            samples.append(current_gmm)
        return samples

class AnnealedImportanceSampling(MonteCarloBase):
    def __init__(self, proposal, priors, betas):
        """
        Base algorithm for Annealed Importance Sampling.

        Parameters
        ----------
        proposal : Proposal Object
            The proposal function to be used.
        prior : GMMPrior Object
            The prior distribution on the parameters of the GMM.
        betas : 1-D array_like of shape (n_betas,)
            Betas used for the annealing process. A series of doubles from 0 to 1.

        References
        ----------
        Neal, Radford M. "Annealed importance sampling." Statistics and Computing 11.2 (2001): 125-139.
        """
        betas = np.array(betas)
        self.targets = [GMMPosteriorTarget(priors, beta,) for beta in betas]
        self.proposal = proposal
        self.priors = priors

    def sample(self, X, n_samples, n_jobs=1, diagnostics=None):
        """
        Generate samples from the posterior distribution of the parameters.

        Parameters
        ----------
        X : 2-D array_like of shape (n_feature_vectors, n_features)
            Feature vectors used as the data for the underlying model.
        n_samples : int
            Number of Monte Carlo samples to be drawn from the posterior distribution.
        n_jobs : int
            Number of cpu cores to utilise during the Monte Carlo simulation.
        diagnostics : empty dictionary, optional
            If included, the dictionary passed to the function will contain diagnostic information
            from each annealing run of AIS. TODO: expand on this

        Returns
        -------
            : list of tuples (GMM, double)
            A list of GMM samples with their corresponding weight.
        """
        samples = []
        for run in xrange(n_samples):
            logging.info('Run: {0}'.format(run))
            if diagnostics is not None:
                diagnostics[run] = {}
                sample, weight = self.anneal(X, n_jobs, diagnostics[run])
            else:
                sample, weight = self.anneal(X, n_jobs)
            samples.append((sample, weight))

        return samples

    def anneal(self, X, n_jobs, diagnostics=None):
        """
        A single annealing run from AIS.

        Parameters
        ----------
        X : 2-D array_like of shape (n_feature_vectors, n_features)
            Feature vectors used as the data for the underlying model.
        n_jobs : int
            Number of cpu cores to utilise during the Monte Carlo simulation.
        diagnostics : empty dictionary, optional
            If included, the dictionary passed to the function will contain diagnostic information
            from each annealing run of AIS. TODO: expand on this

        Returns
        -------
            : tuple (GMM Object, double)
            A single GMM sample from AIS and its corresponding weight.
        """
        # draw from prior
        cur_gmm = self.priors.sample()
        samples = []
        for anneal_run, target in enumerate(self.targets[1:-1]):
            # skip first T_n (prior only) and last transition T_0 (posterior) (not necessary)
            samples.append(cur_gmm)
            cur_gmm = self.proposal.propose(X, cur_gmm, target, n_jobs)

        samples.append(cur_gmm)

        if diagnostics is not None:
            diagnostics['intermediate_log_weights'] = []
            diagnostics['intermediate_betas'] = []

        numerator = 0
        denominator = 0
        for run, sample in enumerate(samples):
            numerator += self.targets[run + 1].log_prob(X, sample, n_jobs)
            denominator += self.targets[run].log_prob(X, sample, n_jobs)
            if diagnostics is not None:
                diagnostics['intermediate_log_weights'].append(numerator - denominator)
                diagnostics['intermediate_betas'].append(self.targets[run].beta)

        weight = numerator - denominator

        if diagnostics is not None:
            diagnostics['intermediate_samples'] = samples
            diagnostics['final_sample'] = cur_gmm
            diagnostics['final_weight'] = weight

        return (cur_gmm, weight)




from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy as np
import os

sourcefiles = [ 'fast_likelihood.pyx']
ext_modules = [Extension("fast_likelihood",
                          sourcefiles,
                          include_dirs = [np.get_include()],
                          extra_compile_args=['-O3', '-fopenmp', '-lc++'],
                          extra_link_args=['-fopenmp'],
                          language='c++')]

setup(
  name = 'fastgmm',
  cmdclass = {'build_ext': build_ext},
  ext_modules = ext_modules
)


from unittest import TestCase

from gmmmc.priors.prior import *


class TestGMMPrior(TestCase):
    def setUp(self):
        self.covars = np.array([[3, 5], [2, 1]])
        means = np.array([[1.0, 2], [3, 4]])
        self.weights = np.array([0.3, 0.7])
        prior_width = 1.0 * np.ones((2, 2))
        self.prior = GMMPrior(MeansGaussianPrior(means, prior_width),
                              CovarsStaticPrior(self.covars), WeightsStaticPrior(self.weights))

    def test_log_prob_gaussian_means(self):
        test_gmm = GMM(np.array([[4, 5], [5, 3]]), self.covars, self.weights)
        test = self.prior.log_prob(test_gmm)
        real = -15.175754132818689
        self.assertAlmostEqual(test, real)

    def test_sample(self):
        pass


class TestMeansGaussianPrior(TestCase):
    def setUp(self):
        means = np.array([[1.0, 2], [3, 4]])
        prior_width = 1.0 * np.ones((2, 2))
        self.prior = MeansGaussianPrior(means, prior_width)

    def test_log_prob(self):
        test = self.prior.log_prob(np.array([[4, 5], [5, 3]]))
        real = -15.175754132818689
        self.assertAlmostEqual(test, real)

    def test_cache_log_prob(self):
        test = self.prior.log_prob(np.array([[4, 5], [5, 3]]))
        real = -15.175754132818689
        self.assertAlmostEqual(test, real)
        test = self.prior.log_prob(np.array([[4, 5], [5, 3]]))
        real = -15.175754132818689
        self.assertAlmostEqual(test, real)
        test = self.prior.log_prob(np.array([[4, 5], [5, 3]]))
        real = -15.175754132818689
        self.assertAlmostEqual(test, real)

    def test_sample(self):
        pass


class TestMeansUniformPrior(TestCase):
    def setUp(self):
        self.prior = MeansUniformPrior(-1.0, 1.0, n_mixtures=5, n_features=10)

    def test_log_prob(self):
        means = np.ones((5, 10))
        self.assertEqual(self.prior.log_prob(means), 0.0)
        self.assertEqual(self.prior.log_prob(means * 2), -np.inf)
        means[3][2] = -10
        self.assertEqual(self.prior.log_prob(means), -np.inf)

    def test_log_prob_single(self):
        mean = np.ones((10))
        self.assertEqual(self.prior.log_prob_single(mean, 3), 0.0)
        self.assertEqual(self.prior.log_prob_single(mean * 2, 4), -np.inf)
        mean[3] = -10
        self.assertEqual(self.prior.log_prob_single(mean, 3), -np.inf)

    def test_sample(self):
        for i in xrange(100):
            sample = self.prior.sample()
            self.assertTrue((sample > -1.0).all() and (sample < 1.0).all())


class TestDiagCovarsUniformPrior(TestCase):
    def setUp(self):
        self.prior = DiagCovarsUniformPrior(0.001, 1.0, 5, 10)

    def test_log_prob(self):
        covars = np.ones((5, 10)) * 0.5
        self.assertEqual(self.prior.log_prob(covars), 0.0)
        self.assertEqual(self.prior.log_prob(covars * 4), -np.inf)
        covars[3][2] = -10
        self.assertEqual(self.prior.log_prob(covars), -np.inf)

    def test_log_prob_single(self):
        covar = np.ones((10))
        self.assertEqual(self.prior.log_prob_single(covar, 3), 0.0)
        self.assertEqual(self.prior.log_prob_single(covar * 2, 4), -np.inf)
        covar[3] = -10
        self.assertEqual(self.prior.log_prob_single(covar, 3), -np.inf)

    def test_sample(self):
        for i in xrange(100):
            sample = self.prior.sample()
            self.assertTrue((sample >= 0.001).all() and (sample <= 1.0).all())


class TestWeightsUniformPrior(TestCase):
    def setUp(self):
        self.prior = WeightsUniformPrior(10)

    def test_log_prob(self):
        weights = np.array([0.1 for i in xrange(10)])
        self.assertEqual(self.prior.log_prob(weights), 12.801827480081469)

    def test_sample(self):
        for _ in xrange(100):
            sample = self.prior.sample()
            self.assertTrue((sample > 0.0).all())
            self.assertAlmostEqual(np.sum(sample), 1.0)




import numpy as np
import sklearn.mixture
from gmmmc.fastgmm import gmm_likelihood
import multiprocessing

class GMM():

    def __init__(self, means, covariances, weights):
        """
        Gaussian Mixture Model Distribution class for calculation of log likelihood and sampling.

        Parameters
        ----------
        means : 2-D array_like of shape (n_mixtures, n_features)
            Means for each component of the GMM
        covariances : 2-D array_like of shape (n_mixtures, n_features)
            Covariance matrices of the GMM. Only diagonal matrices are supported at this time.
        weights : 1-D array_like of shape (n_mixtures,)
            Weights for each of the GMM components
        """
        if len(covariances.shape) == 2:
            self.covariance_type = 'diag'
        else:
            raise NotImplementedError('Only diagonal covariance matrices supported')
        self.gmm = sklearn.mixture.GMM(n_components=len(weights))
        self.gmm.weights_ = weights
        self.gmm.covars_ = covariances
        self.gmm.means_ = means
        self.n_mixtures = len(weights)
        try:
            self.n_features = means.shape[1]
        except:
            raise ValueError("Means array must be 2 dimensional")

    @property
    def means(self):
        return self.gmm.means_

    @property
    def covars(self):
        return self.gmm.covars_

    @property
    def weights(self):
        return self.gmm.weights_

    @means.setter
    def means(self, means):
        # must create GMM object again so that the sklearn sample method will work correctly
        self.gmm = create_gmm(self.n_mixtures, means, self.gmm.covars_, self.gmm.weights_)

    @covars.setter
    def covars(self, covars):
        self.gmm = create_gmm(self.n_mixtures, self.gmm.means_, covars, self.gmm.weights_)

    @weights.setter
    def weights(self, weights):
        self.gmm = create_gmm(self.n_mixtures, self.gmm.means_, self.gmm.covars_, weights)

    def sample(self, n_samples):
        """
        Sample from the GMM.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw.

        Returns
        -------
            : 2-D array_like of shape (n_samples, n_features)
            Samples drawn from the GMM distribution
        """
        return self.gmm.sample(n_samples)

    def log_likelihood(self, X, n_jobs=1):
        """
        Calculate the average log likelihood of the data given the GMM parameters

        Parameters
        ----------
        X : 2-D array_like of shape (n_samples, n_features)
            Data to be used.
        n_jobs : int
            Number of CPU cores to use in the calculation

        Returns
        -------
            : float
            average log likelihood of the data given the GMM parameters

        Notes
        -------
        For GMMs with small numbers of mixtures (<10) the use of more than 1 core can slow down the function.
        """
        n_samples = X.shape[0]
        if n_jobs == 0:
            raise ValueError("n_jobs==0 has no meaning")
        elif n_jobs < 0:
            n_jobs = multiprocessing.cpu_count()
        else:
            n_jobs = n_jobs

        if n_jobs == 1:
            # Use the sklearn/numpy implementation
            return np.sum(self.gmm.score(X))
        else:
            # Sse compiled multireaded C code
            return gmm_likelihood(X, self.means, self.covars, self.weights, n_jobs=n_jobs)




import time
from unittest import TestCase
import numpy as np
import sklearn.mixture
from gmmmc.gmm import GMM


class TestGMM(TestCase):

    def setUp(self):
        self.n_mixtures, self.n_features = 32, 16
        means = np.array(np.random.uniform(-1,1,size=(self.n_mixtures, self.n_features)), dtype=np.float64)
        weights = np.array(np.random.dirichlet([1 for _ in xrange(self.n_mixtures)]), dtype=np.float64)
        covars = np.array(np.random.uniform(0,1,size=(self.n_mixtures, self.n_features)), dtype=np.float64)
        self.gmm = GMM(means, covars, weights)
        self.true_gmm = sklearn.mixture.GMM(n_components=self.n_mixtures)
        self.true_gmm.means_ = means
        self.true_gmm.covars_ = covars
        self.true_gmm.weights_ = weights

    def test_sample(self):
        self.assertTrue(True)

    def test_log_likelihood(self):
        data = np.array(np.random.uniform(-1, 1, (2, self.n_features)), dtype=np.float64)
        start = time.time()
        mine = self.gmm.log_likelihood(data, 1)
        print time.time() - start
        start = time.time()
        sklearns = np.sum(self.true_gmm.score(data))
        print time.time() - start
        self.assertAlmostEqual(mine, sklearns)

    def test_scaling(self):
        store_C = []
        store_sklearn = []
        suite = [10, 100, 1000, 10000, 100000, 1000000]
        for n_samples in suite:
            data = np.array(np.random.uniform(-1, 1, (n_samples, self.n_features)), dtype=np.float64)
            start = time.time()
            mine = self.gmm.log_likelihood(data, -1)
            finish_C = time.time() - start
            store_C.append(str(finish_C))
            start = time.time()
            sklearns = np.sum(self.true_gmm.score(data))
            finish_sklearn = time.time() - start
            store_sklearn.append(str(finish_sklearn))
            self.assertAlmostEqual(mine, sklearns, places=3)




        print 'C'
        print '\n'.join(store_C)
        print 'sklearn'
        print '\n'.join(store_sklearn)



import numpy as np
from gmmmc.gmm import GMM
from gmmmc.proposals.proposals import Proposal
import pdb

class GaussianStepMeansProposal(Proposal):
    """Gaussian Proposal distribution for means of a GMM"""

    def __init__(self, step_sizes=(0.001,)):
        """
        Gaussian proposal distribution for the means. The multivariate Gaussian is centered at the means of the current
        state in the Markov Chain and has covariance given by step_sizes. Multiple step sizes can be specified.
        The proposal algorithm will take these steps in the sequence specified in step_sizes.

        Parameters
        ----------
        step_sizes : 1-D array_like
            Iterable containing the sequence of step sizes (covariances of the Gaussian proposal distribution"
        """
        super(GaussianStepMeansProposal, self).__init__()
        self.step_sizes = step_sizes
        self.count_accepted = np.zeros((len(step_sizes),))
        self.count_illegal =  np.zeros((len(step_sizes),))
        self.count_proposed = np.zeros((len(step_sizes),))

    def propose(self, X, gmm, target, n_jobs=1):
        """
        Propose a new set of GMM means.

        Parameters
        ----------
        X : 2-D array like of shape (n_samples, n_features)
            The observed data or evidence.
        gmm : GMM object
            The current state (set of gmm parameters) in the Markov Chain
        target : GMMPosteriorTarget object
            The target distribution to be found.
        n_jobs : int
            Number of cpu cores to use in the calculation of log probabilities.

        Returns
        -------
            : GMM
            A new GMM object initialised with new mean parameters.
        """
        new_means = np.array(gmm.means)
        beta = target.beta
        prior = target.prior
        steps = [np.random.multivariate_normal(np.zeros(gmm.n_features),
                                              step_size * np.eye(gmm.n_features),
                                              size=gmm.n_mixtures)
                 for step_size in self.step_sizes]

        # calculation of prior probabilities of only the means, since only means will change
        log_priors = np.array([prior.means_prior.log_prob_single(gmm.means[mixture], mixture) for mixture in xrange(gmm.n_mixtures)])
        log_prob_priors = np.sum(log_priors)
        previous_prob = beta * gmm.log_likelihood(X, n_jobs) + np.sum(log_priors)
        for i, step in enumerate(steps):
            for mixture in xrange(gmm.n_mixtures):
                self.count_proposed[i] += 1
                # propose new means
                new_mixture_means = gmm.means[mixture] + step[mixture]

                # try out the new means
                proposed_means = np.array(new_means)
                proposed_means[mixture] = new_mixture_means
                proposed_gmm = GMM(proposed_means, np.array(gmm.covars), np.array(gmm.weights))

                # calculate new prior
                new_log_prob_mixture = prior.means_prior.log_prob_single(new_mixture_means, mixture)
                new_log_prob_priors = log_prob_priors - log_priors[mixture] + new_log_prob_mixture

                # priors
                proposed_prob = beta * proposed_gmm.log_likelihood(X, n_jobs) + new_log_prob_priors
                # ratio

                ratio = proposed_prob - previous_prob
                if ratio > 0 or ratio > np.log(np.random.uniform()):
                    # accept proposal
                    new_means = proposed_means
                    previous_prob = proposed_prob
                    # update prior probability calculation
                    log_prob_priors = new_log_prob_priors
                    log_priors[mixture] = new_log_prob_mixture
                    self.count_accepted[i] += 1

        return GMM(new_means, np.array(gmm.covars), np.array(gmm.weights))

class GaussianStepCovarProposal(Proposal):

    def __init__(self, step_sizes=(0.001,)):
        """
        Gaussian proposal function for the covariances of the GMM.

        Parameters
        ----------
        step_sizes : array_like
            Array of covariance values for the Gaussian proposal.
        """
        super(GaussianStepCovarProposal, self).__init__()
        self.step_sizes = step_sizes
        self.count_accepted = np.zeros((len(step_sizes),))
        self.count_illegal =  np.zeros((len(step_sizes),))
        self.count_proposed = np.zeros((len(step_sizes),))

    def propose(self, X, gmm, target, n_jobs=1):
        """
        Propose a new set of GMM covariances (diagonal only).

        Parameters
        ----------
        X : 2-D array like of shape (n_samples, n_features)
            The observed data or evidence.
        gmm : GMM object
            The current state (set of gmm parameters) in the Markov Chain
        target : GMMPosteriorTarget object
            The target distribution to be found.
        n_jobs : int
            Number of cpu cores to use in the calculation of log probabilities.

        Returns
        -------
            : GMM
            A new GMM object initialised with new covariance parameters.
        """
        new_covars = np.array(gmm.covars)
        beta = target.beta
        prior = target.prior
        previous_prob = beta * gmm.log_likelihood(X, n_jobs) + prior.log_prob(gmm)
        steps = [np.random.multivariate_normal(np.zeros(gmm.n_features),
                                               step_size * np.eye(gmm.n_features),
                                               size=gmm.n_mixtures) for step_size in self.step_sizes]

        log_priors = np.array([prior.means_prior.log_prob_single(gmm.means[mixture], mixture) for mixture in xrange(gmm.n_mixtures)])
        log_prob_priors = np.sum(log_priors)
        for i, step in enumerate(steps):
            for mixture in xrange(gmm.n_mixtures):
                self.count_proposed[i] += 1
                # propose new covars
                new_mixture_covars = gmm.covars[mixture] + step[mixture]

                if (new_mixture_covars > 0).all(): # check covariances are valid
                    # try out the new covars
                    proposed_covars = np.array(new_covars)
                    proposed_covars[mixture] = new_mixture_covars
                    proposed_gmm = GMM(np.array(gmm.means), proposed_covars, np.array(gmm.weights))

                    # calculate desired distribution
                    new_log_prob_mixture = prior.covars_prior.log_prob_single(new_mixture_covars, mixture)
                    new_log_prob_priors = log_prob_priors - log_priors[mixture] + new_log_prob_mixture
                    proposed_prob = beta * proposed_gmm.log_likelihood(X, n_jobs) + new_log_prob_priors

                    # ratio
                    ratio = proposed_prob - previous_prob
                    if ratio > 0 or ratio > np.log(np.random.uniform()):
                        # accept proposal
                        new_covars = proposed_covars
                        previous_prob = proposed_prob
                        log_prob_priors = new_log_prob_priors
                        log_priors[mixture] = new_log_prob_mixture
                        self.count_accepted[i] += 1
                else:
                    self.count_illegal[i] += 1

        return GMM(np.array(gmm.means), np.array(new_covars), np.array(gmm.weights))

class GaussianStepWeightsProposal(Proposal):

    def __init__(self,  n_mixtures, step_sizes=(0.001,)):
        """
        Gaussian proposal function for the weights of a GMM.

        Parameters
        ----------
        n_mixtures
        step_sizes

        Notes
        ----------
        The proposal function works by projecting the weight vector w onto the simplex defined by
        w_1 + w_2 + ..... w_n = 1 , 0<=w_i<=1. The change of basis matrix is found by finding n-1 vectors lying on the plane
        and using gramm schmidt to get an orthonormal basis. A Gaussian proposal function in (n-1)-d space is
        used to find the next point on the simplex.
        """
        super(GaussianStepWeightsProposal, self).__init__()
        self.step_sizes = step_sizes
        self.n_mixtures = n_mixtures
        self.count_accepted = np.zeros((len(step_sizes),))
        self.count_illegal =  np.zeros((len(step_sizes),))
        self.count_proposed = np.zeros((len(step_sizes),))


        if n_mixtures > 1:
            # get change of basis matrix mapping n dim coodinates to n-1 dim coordinates on simplex
            # x1 + x2 + x3 ..... =1
            points = np.random.dirichlet([1 for i in xrange(n_mixtures)], size=n_mixtures - 1)
            points = points.T
            self.plane_origin = np.ones((n_mixtures)) / float(n_mixtures)
            # get vectors parallel to plane from its center (1/n,1/n,....)
            parallel = points - np.ones(points.shape) / float(n_mixtures)
            # do gramm schmidt to get mutually orthonormal vectors (basis)
            self.e, _ = np.linalg.qr(parallel)

    def transformSimplex(self, weights):
        """
        Project weight vector onto the normal simplex.

        Parameters
        ----------
        weights : array_like of shape (n_mixtures,)
            vector of weights for each gaussian component

        Returns
        -------
            : array_like of shape (n_mixtures-1,)
            vector of weights projected onto the simplex plane
        """
        # project onto the simplex
        return np.dot(self.e.T, weights - self.plane_origin)

    def invTransformSimplex(self, simplex_coords):
        """
        Transforms a point on the simplex to the original vector space.

        Parameters
        ----------
        simplex_coords : array_like of shape (n_mixtures - 1,)
            Coordinates of a weight vector on the simplex.

        Returns
        -------
            : array_like of shape(n_mixtures,)
            vector of weights.
        """
        return self.plane_origin + np.dot(self.e, simplex_coords)

    def propose(self, X, gmm, target, n_jobs=1):
        """
        Propose a new set of weight vectors.
        Parameters
        ----------
        X : 2-D array like of shape (n_samples, n_features)
            The observed data or evidence.
        gmm : GMM object
            The current state (set of gmm parameters) in the Markov Chain
        target : GMMPosteriorTarget object
            The target distribution to be found.
        n_jobs : int
            Number of cpu cores to use in the calculation of log probabilities.
        Returns
        -------
            : GMM
            A new GMM object initialised with new covariance parameters.
        """
        accepted = False
        cur_gmm = gmm
        if gmm.n_mixtures > 1:
            for i, step_size in enumerate(self.step_sizes):
                self.count_proposed[i] += 1
                current_weights_transformed = self.transformSimplex(cur_gmm.weights)
                proposed_weights_transformed = np.random.multivariate_normal(current_weights_transformed,
                                                                             np.eye(self.n_mixtures - 1) * step_size)
                proposed_weights = self.invTransformSimplex(proposed_weights_transformed   )
                if np.logical_and(0 <= proposed_weights, proposed_weights <= 1).all()\
                    and np.isclose(np.sum(proposed_weights), 1.0):
                    previous_prob = target.log_prob(X, cur_gmm, n_jobs)
                    proposed_gmm = GMM(np.array(cur_gmm.means), np.array(cur_gmm.covars), proposed_weights)
                    proposed_prob = target.log_prob(X, proposed_gmm, n_jobs)
                    ratio = proposed_prob - previous_prob
                    if ratio > 0 or ratio > np.log(np.random.uniform()):
                        # accept proposal
                        self.count_accepted[i] += 1
                        accepted = True
                        cur_gmm = proposed_gmm
                else:
                    self.count_illegal[i] += 1

        if accepted is True:
            return cur_gmm
        else:
            return GMM(np.array(gmm.means), np.array(gmm.covars), np.array(gmm.weights))




import abc
import numpy as np
import scipy.stats
from gmmmc import GMM

class GMMPrior():
    def __init__(self, means_prior, covars_prior, weights_prior):
        """
        Containing class for the priors of a GMM. Initialised with priors of means, weights & covariances.

        Parameters
        ----------
        means_prior : GMMParameterPrior
            Prior for the means of the GMM
        covars_prior : GMMParameterPrior
            Prior for the covariances of the GMM
        weights_prior : GMMParameterPrior
            Prior for the weights of the GMM
        """
        self.means_prior = means_prior
        self.covars_prior = covars_prior
        self.weights_prior = weights_prior

    def log_prob(self, gmm):
        """

        Parameters
        ----------
        gmm : GMM
            object containing parameters

        Returns
        -------
            : double
            log prior probability of the parameters of the input GMM
        """

        logprob = 0
        logprob += self.means_prior.log_prob(gmm.means)
        logprob += self.covars_prior.log_prob(gmm.covars)
        logprob += self.weights_prior.log_prob(gmm.weights)
        return logprob

    def sample(self):
        """
        Compute a sample from the prior distribution of the GMM's parameters.

        Returns
        -------
            : GMM
            GMM object containing the sampled parameters
        """
        return GMM(self.means_prior.sample(), self.covars_prior.sample(), self.weights_prior.sample())

class GMMParameterPrior:

    @abc.abstractmethod
    def log_prob(self, params):
        """
        Calculate the log prior probability.

        Parameters
        ----------
        params : array_like of varying shape
            parameters for a GMM e.g means weights covariances

        Returns
        -------
            : double
            log prior probability of the parameters

        """
        pass

    @abc.abstractmethod
    def log_prob_single(self, param, mixture_num):
        """
        Compute log probability of a single set of parameters (mean/covariance/weight) vector

        Parameters
        ----------
        param : 1-D array_like vector of parameters for a single mixture
        mixture_num : Mixture index for the parameters.

        Returns
        -------
            : double
            log prior probability of parameters
        """

        pass

class MeansGaussianPrior(GMMParameterPrior):
    def __init__(self, prior_means, covariances):
        """
        Multivariate Gaussian prior distribution for GMM means.
        Each mean vector for the means has its own Gaussian prior. Diagonal covariances for the means.

        Parameters
        ----------
        prior_means : 2-D array_like of shape (n_mixtures, n_features)
            Expected means for each mixture of the GMM
        covariances : 2-D array_like of shape (n_mixtures, n_features)
            Diagonal covariances of the expected means of the GMM. Alters the 'width' of the prior.
        """
        # shape should be (n_mixtures, n_features)
        self.means = prior_means
        self.covars = covariances
        self.distributions = [scipy.stats.multivariate_normal(self.means[i], self.covars[i])\
                              for i in xrange(self.means.shape[0])]
        try:
            self.n_features = prior_means.shape[1]
        except:
            raise ValueError("Means must be 2d")

    def log_prob(self, means):
        """Compute the log prior probability of the means of a GMM according to Gaussian priors.

        Parameters
        ----------
        means : 2-D array_like, of shape (n_mixtures, n_features)
            mean vectors for the GMM

        Returns
        -------
            : double
            Proportional to the log probability of the means of the GMM
        """
        log_prob = np.sum([self.log_prob_single(means[i], i) for i in xrange(len(self.distributions))])

        return log_prob

    def log_prob_single(self, mean, mixture_num):
        """Compute the log probability of the means for a specific mixture.

        Parameters
        ----------
        mean : 1-D array_like of length n_features
            Single mean vector from a single mixture of the GMM.
        mixture_num : int
            Index of the mixture for the mean.

        Returns
        -------
            : double
            Proportional to the log prior probability for means
        """

        return self.distributions[mixture_num].logpdf(mean)

    def sample(self):
        """
        Draw a sample from the Gaussian prior distributions of the mean vectors.

        Returns
        -------
            : 2-D array_like of shape (n_mixtures, n_features)
            Return a complete set of mean vectors for a GMM
        """
        return np.array([[normal.rvs()] if self.n_features == 1 else normal.rvs() for normal in self.distributions])

class MeansUniformPrior(GMMParameterPrior):
    def __init__(self, low, high, n_mixtures, n_features):
        """
        Uniform prior for the means of a GMM.

        Parameters
        ----------
        low : double
            Upper limit for uniform distribution.
        high : double
            Lower limit for uniform distribution.
        n_mixtures : int
            Number of mixtures in GMM.
        n_features : int
            Dimensionality of the feature space.
        """
        self.low = low
        self.high = high
        self.n_mixtures = n_mixtures
        self.n_features = n_features

    def log_prob(self, means):
        """
        Compute the log prior probability of the means of a GMM. Since this will be used for monte carlo simulations,
        we care only that it is proportional to the true probability. For a uniform distribution we can
        use any value.

        Parameters
        ----------
        means : 2-D array_like, of shape (n_mixtures, n_features)
            mean vectors for the GMM

        Returns
        -------
            : double
            Proportional to the log probability of the means of the GMM.
            0.0 if the means are within the bounds of the uniform prior, -inf otherwise.
        """
        # only care about proportional.
        if (means < self.low).any() or (means > self.high).any():
            return -np.inf
        else:
            return 0.0

    def log_prob_single(self, mean, mixture):
        """
        Compute the log probability of the means for a specific mixture.

        Parameters
        ----------
        mean : 1-D array_like of length n_features
            Single mean vector from a single mixture of the GMM.
        mixture_num : int
            Index of the mixture for the mean.

        Returns
        -------
            : double
            Proportional to the log prior probability for means
            0.0 if the means are within the bounds of the uniform prior, -inf otherwise.
        """
        if (mean < self.low).any() or (mean > self.high).any():
            # if invalid value
            return -np.inf
        else:
            return 0.0

    def sample(self):
        # sample means
        return np.random.uniform(self.low, self.high, size=(self.n_mixtures, self.n_features))

class DiagCovarsUniformPrior(GMMParameterPrior):
    def __init__(self, low, high, n_mixtures, n_features):
        """
        Uniform prior for a diagonal covariance matrix in a GMM.

        Parameters
        ----------
        low : double
            Lower limit for uniform distribution. Must be greater than 0.
        high : double
            Upper limit for uniform distribution.
        n_mixtures : int
            Number of mixtures in GMM.
        n_features : int
            Dimensionality of the feature space.
        """
        self.low = low
        self.high = high
        self.n_mixtures = n_mixtures
        self.n_features = n_features

    def log_prob(self, covars):
        """
        Compute the log prior probability of the covariances of a GMM.
        Since this will be used for monte carlo simulations, we care only that it is proportional to
        the true probability. For a uniform distribution we can use any value.

        Parameters
        ----------
        covars : 2-D array_like, of shape (n_mixtures, n_features)
            covariance vectors for the GMM

        Returns
        -------
            : double
            Proportional to the log probability of the covariance of the GMM.
            0.0 if the means are within the bounds of the uniform prior, -inf otherwise.eans
        """

        if (covars < self.low).any() or (covars > self.high).any():
            return -np.inf
        else:
            # return some constant value
            return 0.0

    def log_prob_single(self, covar, mixture_num):
        """
        Compute the log probability of the covariances for a specific mixture.

        Parameters
        ----------
        covar : 1-D array_like of length n_features
            Single diagonal covariance from a single mixture of the GMM.
        mixture_num : int
            Index of the mixture for the covariance matrix.

        Returns
        -------
        double
            Proportional to the log prior probability for the covariance.
            0.0 if the means are wimeansthin the bounds of the uniform prior, -inf otherwise.
        """
        if (covar < self.low).any() or (covar > self.high).any():
            return -np.inf
        else:
            # return some constant value
            return 0.0

    def sample(self):
        """
        Draw a sample for the diagonals of a covariance matrix assuming a uniform prior over each individual element.

        Returns
        -------
            : 2-D array_like of shape (n_mixtures, n_features)
            array of diagonal covariances for a GMM.
        """
        return np.random.uniform(self.low, self.high, size=(self.n_mixtures, self.n_features))

class CovarsStaticPrior(GMMParameterPrior):
    def __init__(self, prior_covars):
        """
        Prior for covariance assuming that we know its true value.
        i.e a probability of 1 at prior_covars and 0 otherwise.

        Parameters
        ----------
        prior_covars : 2-D array_like of shape (n_mixtures, n_features)
            Assumed true values of covariance matrices for GMM
        """
        self.prior_covars = prior_covars

    def log_prob(self, covariances):
        """
        Log probability of a covariance given we know what its true value 'should' be.

        Parameters
        ----------
        covariances : covariance matrices of GMM distribution

        Returns
        -------
            : double
            0 if covariances are identical to their true values, -inf otherwise.
        """
        # assign probability of 1 to the static covariance prior
        if np.allclose(self.prior_covars, covariances):
            return 0.0
        else:
            return -np.inf

    def log_prob_single(self, covariance, mixture_num):
        """
        Log probability of a covariance matrix of a single mixture given we know what its true value should be.

        Parameters
        ----------
        covariance : 1-D array_like of shape (n_features)
            covariance matrix for a specific mixture in GMM
        mixture_num : int
            Index of mixture in GMM

        Returns
        -------
            : double
            0 if covariance is identical to its true value, -inf otherwise
        """
        if np.allclose(self.prior_covars[mixture_num], covariance):
            return 0.0
        else:
            return -np.inf

    def sample(self):
        return np.array(self.prior_covars)

class WeightsUniformPrior(GMMParameterPrior):
    def __init__(self, n_mixtures):
        """
        Uniform Prior for Weights (Dirichlet distribution with all parameters equal to 1)

        Parameters
        ----------
        n_mixtures : int
            number of mixtures in the GMM
        """

        self.alpha = [1 for _ in xrange(n_mixtures)]
        self.dirichlet = scipy.stats.dirichlet(self.alpha)

    def log_prob(self, weights):
        """
        Returns a log probability according to uniform dirichlet prior.

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)
            Weight vector for mixture. Must lie on the normal simplex.

        Returns
        -------
            : double
            log probability under uniform prior.
        """
        return self.dirichlet.logpdf(weights)

    def log_prob_single(self, weights, mixture_num):
        """
        Functionally the same as log_prob

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)
            Weight vector for mixture. Must lie on the normal simplex.
        mixture_num : not used in this context

        Returns
        -------
            : double
            log probability under uniform prior.
        """
        return self.dirichlet.logpdf(weights)

    def sample(self):
        """
        Draw sample from dirichlet distribution

        Returns
        -------
            : 1-D array_like of shape (n_mixtures)
            Set of weight parameters for a GMM.
        """
        return np.random.dirichlet(self.alpha, 1)[0]

class WeightsStaticPrior(GMMParameterPrior):
    def __init__(self, prior_weights):
        """
        Prior for Weights  assuming it has a true fixed value.

        Parameters
        ----------
        n_mixtures : int
            number of mixtures in the GMM
        """
        self.prior_weights = prior_weights

    def log_prob(self, weights):
        """
        Returns a log probability assuming weights have a true fixed value.

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)
            Weight vector for mixture. Must lie on the normal simplex.

        Returns
        -------
            : double
            0 if weights are close to true values, -inf otherwise
        """
        if np.all(np.isclose(weights, self.prior_weights)):
            return 0.0
        else:
            return -np.inf

    def log_prob_single(self, weights, mixture_num):
        """
        Functionally the same as log_prob

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)
            Weight vector for mixture. Must lie on the normal simplex.
        mixture_num : int
            Not used.

        Returns
        -------
            : double
            0 if weights are close to true values, -inf otherwise
        """
        return self.log_prob(weights)

    def sample(self):
        """
        Sample true weight vector

        Returns
        -------
            : 1-D array_like with shape (n_mixtures)
            true weights.
        """
        return np.array(self.prior_weights)

class WeightsDirichletPrior(GMMParameterPrior):
    def __init__(self, alpha):
        """
        Dirichlet prior for the weights of a GMM

        Parameters
        ----------
        alpha : 1-D array_like of shape (n_mixtures)
            Parameters of the dirichlet distribution.

        Notes
        ----------
        Alphas of the dirichlet distribution can be set to the expected weights. (e.g those found by ML estimation)
        """
        self.alpha = alpha

    def log_prob(self, weights):
        """
        Calculate log probability of weight vector according to dirichlet prior.

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)

        Returns
        -------
            : double
            log probability under distribution
        """
        return scipy.stats.dirichlet.logpdf(weights, self.alpha)

    def log_prob_single(self, weights, mixture_num):
        """
        Identical to log_prob

        Parameters
        ----------
        weights : 1-D array_like of shape (n_mixtures)
        mixture_num : unused

        Returns
        -------
            : double
            log probability under distribution
        """
        return self.log_prob(weights)

    def sample(self):
        """
        Sample from dirichlet distribution.

        Returns
        -------
            : 1-D array like of shape (n_mixtures)
            Sample weights from dirichlet distribution.
        """
        return np.random.dirichlet(self.alpha)


import abc

class Proposal(object):

    def __init__(self):
        """
        Base class for a proposal function
        """
        self.count_proposed = 0.0
        self.count_accepted = 0.0
        self.count_illegal = 0.0

    def get_acceptance(self):
        """
        Calculate and return the acceptance rate of the proposal.

        Returns
        -------
            : double
            The acceptance rate of the proposal function
        """
        return self.count_accepted / self.count_proposed

    def get_illegal(self):
        """
        Calculate and return the illegal proposal rate of this proposal.
        (Proposing values outside the support of the parameter space e.g covariances < 0)

        Returns
        -------
            : double
            The illegal proposal rate of the proposal function
        """
        return self.count_illegal / self.count_proposed

    """ generic proposal function"""
    @abc.abstractmethod
    def propose(self, X, gmm, target, n_jobs=1):
        """

        Parameters
        ----------
        X : 2-D array_like
            Observed data or evidence.
        gmm : GMM object

        target
        n_jobs

        Returns
        -------

        """
        pass


class GMMBlockMetropolisProposal(Proposal):

    def __init__(self, propose_mean=None, propose_covars=None, propose_weights=None, propose_iterations=1):
        """
        Blocked Metropolis Proposal function for a GMM.
        The parameters of the means, covariances and weights are updated separately.

        Parameters
        ----------
        propose_mean : Proposal object
            Proposal object
        propose_covars
        propose_weights
        propose_iterations
        """
        self.propose_mean = propose_mean
        self.propose_covars = propose_covars
        self.propose_weights = propose_weights
        self.propose_iterations = propose_iterations

    def propose(self, X, gmm, target, n_jobs=1):
        """
        Propose a new set of gmm parameters. Calls each proposal function one after another.

        Parameters
        ----------
        X : 2-D array_like of shape (n_samples, n_features)
            Feature vectors
        gmm : GMM object
            Current GMM parameters in the markov chain
        target : GMMPosteriorTarget object
            Target distribution
        n_jobs : int
            Number of cpus to use. -1 to use all available cores.

        Returns
        -------
            : GMM Object
            The next state in the Markov Chain.

        """
        new_gmm = gmm
        for _ in xrange(self.propose_iterations):
            if self.propose_mean is not None:
                new_gmm = self.propose_mean.propose(X, new_gmm, target, n_jobs)

            if self.propose_covars is not None:
                new_gmm = self.propose_covars.propose(X, new_gmm, target, n_jobs)

            if self.propose_weights is not None:
                new_gmm = self.propose_weights.propose(X, new_gmm, target, n_jobs)
        return new_gmm



class GMMPosteriorTarget:
    """Posterior distribution (targets distribution)"""
    def __init__(self, prior, beta = 1):
        """
        A posterior target distribution. Calculated by adding the log likelihood and log prior probability.

        Parameters
        ----------
        prior : GMMPrior object
            prior distribution of the GMM parameters
        beta : double
            power of the likelihood component e.g P(X|parameters)^beta * P(parameters). Used for annealing.
        """
        self.prior = prior
        self.beta = beta

    def log_prob(self, X, gmm, n_jobs):
        """

        Parameters
        ----------
        X : 2-D array_like of shape (n_samples, n_features)
            Feature vectors
        gmm : GMM object
            GMM parameters for the calculation of the prior probability
        n_jobs : int
            Number of cores to use in the calculation.

        Returns : double
            log probability of the posterior up to a constant factor.
        -------

        """
        return self.beta * gmm.log_likelihood(X, n_jobs) + self.prior.log_prob(gmm)


import logging
from unittest import TestCase

from scipy.misc import logsumexp

from gmmmc.priors.prior import *
from gmmmc.monte_carlo import MarkovChain
from gmmmc.posterior import GMMPosteriorTarget
from gmmmc.proposals import *


class TestMarkovChain(TestCase):

    def setUp(self):
        np.random.seed(0)
        self.n_mixtures = 4
        self.n_features = 4
        self.n_samples = 1000
        self.truth_gmm = GMM(means=np.random.uniform(low=-1, high=1, size=(self.n_mixtures, self.n_features)),
                             covariances=np.random.uniform(low=0, high=1, size=(self.n_mixtures, self.n_features)),
                             weights=np.random.dirichlet(np.ones(self.n_mixtures)))
        # draw samples from the true distribution
        self.X = self.truth_gmm.sample(self.n_samples)
        logging.getLogger().setLevel(logging.INFO)

    def test_sample(self):
        pass

    def test_simple_means_only(self):
        prior = GMMPrior(MeansUniformPrior(-1, 1, self.n_mixtures, self.n_features),
                         CovarsStaticPrior(self.truth_gmm.covars),
                         WeightsStaticPrior(self.truth_gmm.weights))
        proposal = GMMBlockMetropolisProposal(propose_mean=GaussianStepMeansProposal(step_sizes=[0.003]))
        initial_gmm = GMM(means=np.random.uniform(low=-1, high=1, size=(self.n_mixtures, self.n_features)),
                          covariances=self.truth_gmm.covars, weights=self.truth_gmm.weights)
        mc = MarkovChain(proposal, prior, initial_gmm)
        # make samples
        gmm_samples = mc.sample(self.X, n_samples=10000, n_jobs=-1)
        # discard gmm samples
        gmm_samples = gmm_samples[int(10000 / 2)::50]
        sample = self.X[0]
        mcmc_likelihood = logsumexp([gmm.log_likelihood(np.array([sample]), n_jobs=-1) for gmm in gmm_samples])
        mcmc_likelihood = mcmc_likelihood - np.log(len(gmm_samples))
        true_likelihood = self.truth_gmm.log_likelihood(np.array([sample]))
        logging.info('MCMC Means Acceptance: {0}'.format(proposal.propose_mean.get_acceptance()))
        logging.info('MCMC Likelihood: {0}'.format(str(mcmc_likelihood)))
        logging.info('True Likelihood: {0}'.format(str(true_likelihood)))
        if abs(mcmc_likelihood - true_likelihood) > 2:
            self.fail("MCMC failed to converge")



from distutils.core import setup
from distutils.extension import Extension
import numpy as np
import os
from Cython.Build import cythonize

sourcefiles = ['./gmmmc/fastgmm/fast_likelihood.pyx']
ext_modules = [Extension("fast_likelihood",
                          sourcefiles,
                          include_dirs = [np.get_include()],
                          extra_compile_args=['-O3', '-fopenmp', '-lc++'],
                          extra_link_args=['-fopenmp'],
                          language='c++')]
# ,'./gmmmc/fastgmm/fast_likelihood_threaded.cpp'
setup(
    name='gmmmc',
    version='0.2.1.7',
    packages=['gmmmc', 'gmmmc.tests', 'gmmmc.priors', 'gmmmc.fastgmm', 'gmmmc.proposals'],
    url='http://github.com/jeremy-ma/gmmmc',
    license='',
    author='Jeremy Ma',
    author_email='jeremy.ma@student.unsw.edu.au',
    description='Functions for drawing Monte Carlo samples from GMM parameter space',
    download_url='https://github.com/jeremy-ma/gmmmc/tarball/0.1',
    keywords = ['gmm', 'monte carlo', 'speech'],
    ext_modules= cythonize(ext_modules)
)



from .prior import MeansGaussianPrior, MeansUniformPrior, CovarsStaticPrior, \
                   DiagCovarsUniformPrior, WeightsUniformPrior,\
                   WeightsDirichletPrior, WeightsStaticPrior, GMMPrior
